// channel flow application; works for Reynolds between 0 and 2000

#include <stdio.h>
#include <math.h>
#include <stdlib.h>

typedef double real;

inline real sq(real x){ return x*x; }

real Reynolds = 0.01;
real d = 1;
real a = d/2;
const int factor = 3;
real k = 0.05;					// estimate of entry constants
real h0 = 1;					
real LE = d*sqrt(sq(h0)+sq(k*Reynolds));	// estimate of entry length
real L = factor*LE;

const int NY = 40;
const int NX =  NY*factor;

int poissonmode = 0;

real dx = L/NX;
real dy = a/NY;

real pi = acos(-1.);
real jacobi = (sq(dy)*cos(pi/NX)+sq(dx)*cos(pi/NY))/(sq(dx)+sq(dy));
real omegajacobi = 2/(1+sqrt(1-sq(jacobi)));
real omega = omegajacobi; 
int SORiters = NY;

real u0 = 1;

real nu = (u0*d)/Reynolds;

real tau = 0.2;
real t = 0;
real dt = tau*dx/u0;

real dx2 = dx*dx;
real dy2 = dy*dy;
real dxy2 = dx2*dy2/(dx2+dy2);

int shift = 0;

int iters = 100000;
/*
int write = 1; //1000;
int print = 1; //write/10;
*/
int write = 1000;
int print = write/10;


int ask = 1000;

real accuracy = 0.01;
real precision = 0;
real divergence = 0;
real error = 0;
real total = 0;  

int iter;
int SORiter;

typedef real realmatrix[NX+2][NY+2];

realmatrix VX,VY,P,H;
realmatrix FX,FY,S,G;

char *outname = "channel.data";
FILE *out;


#define DDXC(M,ix,iy) ((M[ix+1][iy]-M[ix-1][iy])/(2*dx))
#define DDXF(M,ix,iy) ((4*M[ix+1][iy]-M[ix+2][iy]-3*M[ix][iy])/(2*dx))
#define DDXB(M,ix,iy) ((4*M[ix-1][iy]-M[ix-2][iy]-3*M[ix][iy])/(-2*dx))
#define D2DX2C(M,ix,iy) ((M[ix+1][iy]+M[ix-1][iy]-2*M[ix][iy])/(sq(dx)))

#define DDYC(M,ix,iy) ((M[ix][iy+1]-M[ix][iy-1])/(2*dy))
#define DDYF(M,ix,iy) ((4*M[ix][iy+1]-M[ix][iy+2]-3*M[ix][iy])/(2*dy))
#define DDYB(M,ix,iy) ((4*M[ix][iy-1]-M[ix][iy-2]-3*M[ix][iy])/(-2*dy))
#define D2DY2C(M,ix,iy) ((M[ix][iy+1]+M[ix][iy-1]-2*M[ix][iy])/(sq(dy)))

#define EXF(M,ix,iy) (2*M[ix+1][iy]-M[ix+2][iy])
#define EXB(M,ix,iy) (2*M[ix-1][iy]-M[ix-2][iy])

#define EYF(M,ix,iy) (2*M[ix][iy+1]-M[ix][iy+2])
#define EYB(M,ix,iy) (2*M[ix][iy-1]-M[ix][iy-2])

// form average relative to a field M, KX and KY can be -1, 0, +1 
#define AVE(M,KX,KY,ix,iy) ((M[ix][iy]+M[ix+KX][iy]+M[ix][iy+KY]+M[ix+KX][iy+KY])/4)


//---------------------------------------------------------------------------------

void Initialize()
{
	int ix,iy;

	// clear all

	for( ix = 0; ix <= NX+1; ix++ )
		for( iy = 0; iy <= NY+1; iy++ )
		{
			VX[ix][iy] = 0;  
			VY[ix][iy] = 0;
			P[ix][iy] = 0; 
			FX[ix][iy] = 0;
			FY[ix][iy] = 0; 
			S[ix][iy] = 0;
		}

	// set up initial field

	for( ix = 0; ix <= NX; ix++)
		for( iy = 0; iy <= NY; iy++)
			VX[ix][iy] = u0;				// bulk

	for( ix = 1; ix <= NX; ix++ )
			VX[ix][NY+1] =  -VX[ix][NY];	// wall

}
//---------------------------------------------------------------------------------

void DoForces()
{
	int ix,iy;
	real vx,vy;
	real dvxdx,dvxdy,dvydx,dvydy;
	real d2vxdx2,d2vxdy2,d2vydx2,d2vydy2;


	// x-force, at internal points

	for( ix = 1; ix <= NX-1; ix++ )
		for( iy = 1; iy <= NY; iy++ )
		{
			vx = VX[ix][iy];
			vy = AVE(VY,1,-1,ix,iy); 
			
			dvxdx = DDXC(VX,ix,iy);
			d2vxdx2 = D2DX2C(VX,ix,iy);

			dvxdy = DDYC(VX,ix,iy);
			d2vxdy2 = D2DY2C(VX,ix,iy);

			FX[ix][iy] = -vx*dvxdx-vy*dvxdy+nu*(d2vxdx2+d2vxdy2);	
		}

	// y-force at internal points

	for( ix = 1; ix <= NX; ix++ )
		for( iy = 1; iy <= NY-1; iy++ )
		{
			vx = AVE(VX,-1,1,ix,iy);
			vy = VY[ix][iy];

			dvydx = DDXC(VY,ix,iy);
			d2vydx2 = D2DX2C(VY,ix,iy);

			dvydy = DDYC(VY,ix,iy);
			d2vydy2 = D2DY2C(VY,ix,iy);

			FY[ix][iy] = -vx*dvydx-vy*dvydy+nu*(d2vydx2+d2vydy2);	
		}

	// boundary forces (only first order)

	if( poissonmode == 0 ) return;					
	for( iy = 1; iy <= NY; iy++ )
	{	
		FX[0][iy] = FX[1][iy];//EXF(FX,1,iy);		// entry
		FX[NX][iy] = FX[NX-1][iy];//EXB(FX,NX+1,iy);  // exit
	}
	for( iy = 1; iy <= NY-1; iy++ )
	{	
		FY[0][iy] = FY[1][iy];//EXF(FY,0,iy);		// entry	
		FY[NX+1][iy] = FY[NX][iy];//EXB(FY,NX+1,iy);	// exit
	}

	for( ix = 0; ix <= NX; ix++ )
	{
		FX[ix][0] = FX[ix][1];			// midline
		FX[ix][NY+1] = FX[ix][NY];//EYB(FX,ix,NY+1);	// wall
	}
	for( ix = 0; ix <= NX+1; ix++ )
	{
		FY[ix][1] = 0;				// midline	
		FY[ix][NY] = FY[ix][NY-1];//EYB(FY,ix,NY+1);	// wall
	}

}
//---------------------------------------------------------------------------------

void DoPressure()
{
	int ix,iy;

	// pressure source

	real norm = 0;
	real divnorm = 0;
	divergence = 0;
	for( ix = 1; ix <= NX; ix++ )
		for( iy = 1; iy <= NY; iy++ )
		{
			real vx = VX[ix][iy];
			real vy = VY[ix][iy];
			real divv = (vx-VX[ix-1][iy])/dx+(vy-VY[ix][iy-1])/dy;
			real divf = (FX[ix][iy]-FX[ix-1][iy])/dx+(FY[ix][iy]-FY[ix][iy-1])/dy;
		    real s = divf+divv/dt;
			S[ix][iy] = s;
			norm += sq(s);
			divergence += sq(divv);
			divnorm += sq(vx)+sq(vy);
		}
	divergence = sqrt(divergence/divnorm);
	norm = sqrt(norm);

	// SOR iterations
	
	for( SORiter = 0; SORiter < SORiters; SORiter++ )
	{
		// set boundary conditions

		for( iy = 1; iy <= NY; iy++ )
		{
			P[0][iy] = P[1][iy]-(( poissonmode >0)? FX[0][iy]*dx : 0);			// entry
			P[NX+1][iy] = P[NX][iy]+ ((poissonmode >0)? FX[NX][iy]*dx : 0);	// exit
		}
		for( ix = 0; ix <= NX+1; ix++ )
		{
			P[ix][0] = P[ix][1];						// midline
			P[ix][NY+1] = P[ix][NY]+((poissonmode >0)? FY[ix][NY]*dy : 0);	// wall
		}

		// perform iteration

		precision = 0;
		for( ix = 1; ix <= NX; ix++)
			for( iy = 1; iy <= NY; iy++)
			{
				real r = D2DX2C(P,ix,iy)+D2DY2C(P,ix,iy)-S[ix][iy];
				H[ix][iy] = r/norm;
				P[ix][iy] += r*omega*dxy2/2;
				precision += r*r;
			}
			precision = ( norm > 0 )? sqrt(precision)/norm :0;
		if( precision < accuracy) break;
	}
	
	// normalize pressure at exit midline

	real p0 = AVE(P,1,1,NX,0);
	for( ix = 0; ix <= NX+1; ix++ )
		for( iy = 0; iy <= NY+1; iy++ )
			P[ix][iy] -= p0;

}

//---------------------------------------------------------------------------------

void DoUpdate()
{	
	int ix,iy;

	// inner x-velocity

	for( ix = 1; ix <= NX-1; ix++ )
		for( iy = 1; iy <= NY; iy++ )
		{
			real vx = VX[ix][iy];
			real dpdx = (P[ix+1][iy]-P[ix][iy])/dx;
			real dvx = dt*(FX[ix][iy]-dpdx);
			VX[ix][iy] += dvx;

			error += sq(dvx);
			total += sq(vx);
		}

	// inner y-velocity
	for( ix = 1; ix <= NX; ix++ )
		for( iy = 1; iy <= NY-1; iy++ )
		{
			real vy = VY[ix][iy];
			real dpdy = (P[ix][iy+1]-P[ix][iy])/dy;
			real dvy = dt*(FY[ix][iy]-dpdy);
			VY[ix][iy] += dvy;
			
			error += sq(dvy);
			total += sq(vy);
		}

	// set boundary values

	for( iy = 1; iy <= NY; iy++ )
	{	
		VX[0][iy] = u0;				// entry
		VX[NX][iy] = VX[NX-1][iy];  // exit
	}
	for( iy = 1; iy <= NY; iy++ )
	{	
		VY[0][iy] = -VY[1][iy];		// entry	
		VY[NX+1][iy] = VY[NX][iy];	// exit
	}

	for( ix = 0; ix <= NX; ix++ )
	{
		VX[ix][0] = VX[ix][1];		// midline
		VX[ix][NY+1] = -VX[ix][NY];	// wall
	}
	for( ix = 0; ix <= NX+1; ix++ )
	{
		VY[ix][0] = 0;				// midline	
		VY[ix][NY] = 0;				// wall
	}

	// new dt

	real vxmax = 0;
	real vymax = 0;
	for( ix = 1 ; ix <= NX-1; ix++)
		for( iy = 1; iy <= NY; iy++)
		{
			real vx = fabs(VX[ix][iy]);
			if( vx > vxmax ) vxmax = vx;
		}
	for( ix = 0 ; ix <= NX+1; ix++)
		for( iy = 0; iy <= NY; iy++)
		{
			real vy = fabs(VY[ix][iy]);
			if( vy > vymax ) vymax = vy;
		}

	dt = tau*__min(dxy2/nu,__min(dx/vxmax,dy/vymax));
}


//---------------------------------------------------------------------------------

void DoCycle()
{
	DoForces();
	DoPressure();
	DoUpdate();
}


//---------------------------------------------------------------------------------

#define realOUTPUT(x) fprintf(out," "#x" = %f;\n",x);
#define intOUTPUT(x) fprintf(out," "#x" = %d;\n",x);

void WriteInfo()
{
	int ix,iy;

	fopen_s(&out,outname,"w");

	// input parameters

	fprintf(out,"{\n");
	intOUTPUT(NX);
	intOUTPUT(NY);

	realOUTPUT(L);
	realOUTPUT(a);
	realOUTPUT(LE);
	realOUTPUT(h0);
	realOUTPUT(nu);
	realOUTPUT(Reynolds);
	realOUTPUT(omega);
	realOUTPUT(omegajacobi);
	realOUTPUT(accuracy);
	realOUTPUT(u0);
	realOUTPUT(dx);
	realOUTPUT(dy);

	intOUTPUT(iters);
		
	
	// x

	fprintf(out," x = {");
	for( ix = 0; ix <= NX; ix++ ) 
	{
		real x = ix*dx;
		if( ix > 0 ) fprintf(out,",");
		fprintf(out,"%f",x);
	}
	fprintf(out,"};\n");

	// y

	fprintf(out," y = {");
	for( iy = 0; iy <= NY; iy++ ) 
	{
		real y = iy*dy;
		if( iy > 0 ) fprintf(out,",");
		fprintf(out,"%f",y);
	}
	fprintf(out,"};\n");

	fprintf(out,"};\n");
	fclose(out);
}

//---------------------------------------------------------------------------------

void WriteRealMatrix(char *name,realmatrix &m, int kx, int ky)
{
	int ix,iy;

	fprintf(out," %s =\n {\n  {",name);
	for( ix = 0; ix <= NX; ix++ )
	{
		if( ix > 0 ) fprintf(out,",\n  {");
		for( iy = 0; iy <= NY; iy++)
		{
			if( iy > 0 ) fprintf(out,",");
			fprintf(out,"%f",AVE(m,1-kx,1-ky,ix,iy));
		}
		fprintf(out,"}");
	}
	fprintf(out,"\n };\n");
}

//---------------------------------------------------------------------------------

void WriteData()
{


	fopen_s(&out,outname,"a+");

	fprintf(out,"{\n");

	intOUTPUT(iter);
	realOUTPUT(precision);
	realOUTPUT(divergence);
	realOUTPUT(t);
	realOUTPUT(dt);

	WriteRealMatrix("vx",VX,1,0);
	WriteRealMatrix("vy",VY,0,1);
	WriteRealMatrix("fx",FX,1,0);
	WriteRealMatrix("fy",FY,0,1);
	WriteRealMatrix("p",P,0,0);
	WriteRealMatrix("h",H,0,0);

	fprintf(out,"};\n");
	fclose(out);
}

//---------------------------------------------------------------------------------

int main()
{
	WriteInfo();
	Initialize();

	printf("\ncycle        t       dt iters prc/acc  div/acc      rate       vx");
	FILE *out1;
	fopen_s(&out1,"timeseq.txt","w");
	fprintf(out1,"timeseq = {");
	for( iter = 0; iter <= iters; iter++ )
	{
		DoCycle();

		if( iter%print == 0 ) 
		{	
			real vx = (VX[NX][0]+VX[NX][1])/2;
			printf("\n%-5d %8.3g %8.2g %4d %8.2g %8.2g %8.2g%% %8.4f", 
				iter,t,dt,SORiter,precision/accuracy,divergence/accuracy,sqrt(error/total)*100,vx);
			total = error = 0;
			if( iter > 0 ) fprintf(out1,",");
			fprintf(out1,"{%f,%f}",t,vx);
		}
		if( iter%write == 0) WriteData();
		t += dt;
		if( iter%ask == 0 ) 
		{	
			int c = getc(stdin);
			if ( c == 'q' ) break;
		}
	}
	fprintf(out1,"};");
	fclose(out1);
	if( iter%write != 0 ) WriteData();
	return 0;
}
